# -*- coding: utf-8 -*-
#
# K2hash Python Driver under MIT License
#
# Copyright (c) 2022 Yahoo Japan Corporation
#
# For the full copyright and license information, please view
# the license file that was distributed with this source code.
#
# AUTHOR:   Hirotaka Wakabayashi
# CREATE:   Tue Feb 08 2022
# REVISION:
#
"""
k2hash package
"""
from __future__ import absolute_import

__all__ = [
    "K2hash",
    "Queue",
    "BaseQueue",
    "KeyQueue",
    "K2hashIterator",
    "LogLevel",
    "KeyPack",
    "AttrPack",
    "OpenFlag",
    "DumpLevel",
    "LogLevel",
    "TimeUnit",
]

import ctypes
import logging
import sys
from ctypes import (
    POINTER,
    Structure,
    c_bool,
    c_char_p,
    c_int,
    c_long,
    c_size_t,
    c_ubyte,
    c_uint64,
    c_ulong,
)
from ctypes.util import find_library
from enum import Enum
from logging import StreamHandler
from logging.handlers import TimedRotatingFileHandler
from typing import List  # noqa: pylint: disable=unused-import
from typing import Dict, Optional, Set, Tuple, Union

LOG = logging.getLogger(__name__)


# https://docs.python.org/3/library/ctypes.html#incomplete-types
class FILE(Structure):  # noqa: pylint:disable=too-few-public-methods
    """C FILE structure"""


# https://docs.python.org/3/library/ctypes.html#incomplete-types
class time_t(Structure):  # noqa: pylint:disable=too-few-public-methods,invalid-name
    """C time_t structure"""


# KeyPack structure
# See: https://github.com/yahoojapan/k2hash/blob/master/lib/k2hash.h#L75
#
# typedef struct k2h_key_pack{
# 	unsigned char*	pkey;
# 	size_t			length;
# }K2HKEYPCK, *PK2HKEYPCK;
#
class KeyPack(Structure):  # noqa: pylint:disable=too-few-public-methods
    """C KeyPack structure"""

    _fields_ = [("pkey", POINTER(c_ubyte)), ("length", c_size_t)]


# AttrPack structure
# See: https://github.com/yahoojapan/k2hash/blob/master/lib/k2hash.h#L81
#
# typedef struct k2h_attr_pack{
# 	unsigned char*	pkey;
# 	size_t			keylength;
# 	unsigned char*	pval;
# 	size_t			vallength;
# }K2HATTRPCK, *PK2HATTRPCK;
#
class AttrPack(Structure):  # noqa: pylint:disable=too-few-public-methods
    """C Attr structure"""

    _fields_ = [
        ("pkey", POINTER(c_ubyte)),
        ("keylength", c_size_t),
        ("pval", POINTER(c_ubyte)),
        ("vallength", c_size_t),
    ]


class OpenFlag(Enum):
    """k2hash file open flags"""

    READ = 1
    EDIT = 2
    TEMPFILE = 3
    MEMORY = 4


class TimeUnit(Enum):
    """k2hash time units"""

    DAYS = 1
    HOURS = 2
    MILLISECONDS = 3
    MINUTES = 4
    SECONDS = 5


class DumpLevel(Enum):
    """k2hash file status information"""

    # Dump headers
    HEADER = 1
    # Dump headers and hash tables
    HASH_TABLE = 2
    # Dump headers, hash tables and sub hash tables
    SUB_HASH_TABLE = 3
    # Dump headers, hash tables, sub hash tables and elements
    ELEMENT = 4
    # Dump headers, hash tables, sub hash tables, elements and pages
    PAGE = 5


class LogLevel(Enum):
    """k2hash log level"""

    # Silent disables logging.
    SILENT = 1
    # logs on errors
    ERROR = 2
    # logs on (errors || warnings)
    WARNING = 3
    # logs on (errors || warnings || info)
    INFO = 4
    # logs on (errors || warnings || info || debug)
    DEBUG = 5


# Library handles
_HANDLE: Dict[str, str] = {}


# Initializes library handles and stores the result in the _HANDLE cache
def _init_library_handle():
    global _HANDLE  # noqa: pylint: disable=global-statement
    if _HANDLE:
        return _HANDLE

    # Loads libc and libk2hash and ...
    result = {}
    result["c"] = _load_libc()
    result["k2hash"] = _load_libk2hash()
    _HANDLE = result

    return result


def _load_libc():
    ret = ctypes.cdll.LoadLibrary(find_library("c"))
    # print("type(ret).{}".format(type(ret)))
    if ret is None:
        raise FileNotFoundError
    return ret


def _load_libk2hash():  # noqa: pylint: disable=too-many-statements
    ret = ctypes.cdll.LoadLibrary(find_library("k2hash"))
    if ret is None:
        return None

    # Defines prototypes for python code
    #
    # 4. find API
    #
    # k2h_find_h k2h_find_first(k2h_h handle)
    ret.k2h_find_first.restype = c_uint64
    ret.k2h_find_first.argtypes = [c_uint64]
    # k2h_find_h k2h_find_first_str_subkey(k2h_h handle, const char* pkey)
    ret.k2h_find_first_str_subkey.restype = c_uint64
    ret.k2h_find_first_str_subkey.argtypes = [c_uint64, c_char_p]
    # k2h_find_h k2h_find_next(k2h_find_h findhandle)
    ret.k2h_find_next.restype = c_uint64
    ret.k2h_find_next.argtypes = [c_uint64]
    # bool k2h_find_get_key(k2h_find_h findhandle, unsigned char** ppkey, size_t* pkeylength)
    ret.k2h_find_get_key.restype = c_uint64
    ret.k2h_find_get_key.argtypes = [c_uint64, POINTER(c_char_p), POINTER(c_size_t)]
    #
    # 3. keyqueue API
    #
    # k2h_keyq_h k2h_keyq_handle_str_prefix(k2h_h handle, bool is_fifo, const char* pref)
    ret.k2h_keyq_handle_str_prefix.restype = c_uint64
    ret.k2h_keyq_handle_str_prefix.argtypes = [c_uint64, c_bool, c_char_p]
    # bool k2h_keyq_str_push_keyval(k2h_keyq_h keyqhandle, const char* pkey, const char* pval)
    ret.k2h_keyq_str_push_keyval.restype = c_bool
    ret.k2h_keyq_str_push_keyval.argtypes = [c_uint64, c_char_p, c_char_p]
    # bool k2h_keyq_str_push_keyval_wa(
    # k2h_keyq_h keyqhandle, const char* pkey, const char* pval, const char* encpass,
    # const time_t* expire)
    ret.k2h_keyq_str_push_keyval_wa.restype = c_bool
    ret.k2h_keyq_str_push_keyval_wa.argtypes = [
        c_uint64,
        c_char_p,
        c_char_p,
        c_char_p,
        POINTER(c_long),
    ]
    # bool k2h_keyq_dump(k2h_keyq_h qhandle, FILE* stream)
    ret.k2h_keyq_dump.restype = c_bool
    ret.k2h_keyq_dump.argtypes = [c_uint64, POINTER(FILE)]
    # bool k2h_keyq_free(k2h_keyq_h qhandle)
    ret.k2h_keyq_free.restype = c_bool
    ret.k2h_keyq_free.argtypes = [c_uint64]
    # int k2h_keyq_count(k2h_keyq_h qhandle)
    ret.k2h_keyq_count.restype = c_int
    ret.k2h_keyq_count.argtypes = [c_uint64]
    # bool k2h_keyq_str_read_keyval_wp(
    # k2h_keyq_h keyqhandle, char** ppkey, char** ppval, int pos, const char* encpass)
    ret.k2h_keyq_str_read_wp.restype = c_bool
    ret.k2h_keyq_str_read_keyval_wp.argtypes = [
        c_uint64,
        POINTER(c_char_p),
        POINTER(c_char_p),
        c_int,
        c_char_p,
    ]
    # bool k2h_keyq_empty(k2h_keyq_h qhandle)
    ret.k2h_keyq_empty.restype = c_bool
    ret.k2h_keyq_empty.argtypes = [c_uint64]
    # bool k2h_keyq_str_pop_keyval_wp(
    # k2h_keyq_h keyqhandle, char** ppkey, char** ppval, const char* encpass)
    ret.k2h_keyq_str_pop_keyval_wp.restype = c_bool
    ret.k2h_keyq_str_pop_keyval_wp.argtypes = [
        c_uint64,
        POINTER(c_char_p),
        POINTER(c_char_p),
        c_char_p,
    ]
    # bool k2h_keyq_remove(k2h_keyq_h qhandle, int count)
    ret.k2h_keyq_remove.restype = c_bool
    ret.k2h_keyq_remove.argtypes = [c_uint64, c_int]

    #
    # 2. queue API
    #
    # k2h_q_h k2h_q_handle_str_prefix(k2h_h handle, bool is_fifo, const char* pref)
    ret.k2h_q_handle_str_prefix.restype = c_uint64
    ret.k2h_q_handle_str_prefix.argtypes = [c_uint64, c_bool, c_char_p]
    # bool k2h_q_str_push_wa(
    #   k2h_q_h qhandle, const char* pdata, const PK2HATTRPCK pattrspck,
    #   int attrspckcnt, const char* encpass, const time_t* expire)
    ret.k2h_q_str_push_wa.restype = c_bool
    ret.k2h_q_str_push_wa.argtypes = [
        c_uint64,
        c_char_p,
        POINTER(AttrPack),
        c_int,
        c_char_p,
        POINTER(c_ulong),
    ]
    # bool k2h_q_str_push(k2h_q_h qhandle, const char* pval)
    ret.k2h_q_str_push.restype = c_bool
    ret.k2h_q_str_push.argtypes = [c_uint64, c_char_p]
    # bool k2h_q_remove(k2h_q_h qhandle, int count)
    ret.k2h_q_remove.restype = c_bool
    ret.k2h_q_remove.argtypes = [c_uint64, c_int]
    # bool k2h_q_free(k2h_q_h qhandle)
    ret.k2h_q_free.restype = c_bool
    ret.k2h_q_free.argtypes = [c_uint64]
    # int k2h_q_count(k2h_q_h qhandle)
    ret.k2h_q_count.restype = c_int
    ret.k2h_q_count.argtypes = [c_uint64]
    # bool k2h_q_str_read_wp(k2h_q_h qhandle, char** ppdata, int pos, const char* encpass)
    ret.k2h_q_read_wp.restype = c_bool
    ret.k2h_q_read_wp.argtypes = [
        c_uint64,
        POINTER(c_char_p),
        POINTER(c_size_t),
        c_int,
        c_char_p,
    ]
    # bool k2h_q_empty(k2h_q_h qhandle)
    ret.k2h_q_empty.restype = c_bool
    ret.k2h_q_empty.argtypes = [c_uint64]
    # bool k2h_q_str_pop(k2h_q_h qhandle, char** ppval)
    # bool k2h_q_str_pop_wp(k2h_q_h qhandle, char** ppdata, const char* encpass)
    ret.k2h_q_str_pop_wp.restype = c_bool
    ret.k2h_q_str_pop_wp.argtypes = [c_uint64, POINTER(c_char_p), c_char_p]
    # bool k2h_q_dump(k2h_q_h qhandle, FILE* stream)
    ret.k2h_q_dump.restype = c_bool
    ret.k2h_q_dump.argtypes = [c_uint64, POINTER(FILE)]

    #
    # 1. k2hash API
    #
    # add attr crypt API
    # bool k2h_add_attr_crypt_pass(k2h_h handle, const char* pass, bool is_default_encrypt)
    ret.k2h_add_attr_crypt_pass.argtypes = [c_uint64, c_char_p, c_bool]
    ret.k2h_add_attr_crypt_pass.restype = c_bool

    # add attr plugin API
    # bool k2h_add_attr_plugin_library(k2h_h handle, const char* libpath)
    ret.k2h_add_attr_plugin_library.argtypes = [c_uint64, c_char_p]
    ret.k2h_add_attr_plugin_library.restype = c_bool

    # add attr API
    # bool k2h_add_str_attr(
    # k2h_h handle, const char* pkey, const char* pattrkey, const char* pattrval)
    ret.k2h_add_str_attr.argtypes = [c_uint64, c_char_p, c_char_p, c_char_p]
    ret.k2h_add_str_attr.restype = c_bool
    # bool k2h_add_attr(
    # k2h_h handle, const unsigned char* pkey, size_t keylength, const unsigned char* pattrkey,
    # size_t attrkeylength, const unsigned char* pattrval, size_t attrvallength)
    ret.k2h_add_attr.argtypes = [
        c_uint64,
        c_char_p,
        c_size_t,
        c_char_p,
        c_size_t,
        c_char_p,
        c_size_t,
    ]
    ret.k2h_add_attr.restype = c_bool

    # add subkey API
    # bool k2h_add_subkey(
    # k2h_h handle, const unsigned char* pkey, size_t keylength, const unsigned char* psubkey,
    # size_t skeylength, const unsigned char* pval, size_t vallength)
    ret.k2h_add_subkey.argtypes = [
        c_uint64,
        c_char_p,
        c_size_t,
        c_char_p,
        c_size_t,
        c_char_p,
        c_size_t,
    ]
    ret.k2h_add_subkey.restype = c_bool
    # bool k2h_add_subkey_wa(k2h_h handle, const unsigned char* pkey, size_t keylength,
    # const unsigned char* psubkey, size_t skeylength, const unsigned char* pval,
    # size_t vallength, const char* pass, const time_t* expire)
    ret.k2h_add_subkey_wa.argtypes = [
        c_uint64,
        c_char_p,
        c_size_t,
        c_char_p,
        c_size_t,
        c_char_p,
        c_size_t,
        c_char_p,
        POINTER(c_ulong),
    ]
    ret.k2h_add_subkey_wa.restype = c_bool

    # close API
    # bool k2h_close(k2h_h handle)
    # bool k2h_close_wait(k2h_h handle, long waitms)
    ret.k2h_close_wait.argtypes = [c_uint64, c_long]
    ret.k2h_close_wait.restype = c_bool

    # create API
    # bool k2h_create(const char* filepath, int maskbitcnt, int cmaskbitcnt,
    # int maxelementcnt, size_t pagesize)
    ret.k2h_create.argtypes = [c_char_p, c_int, c_int, c_int, c_size_t]
    ret.k2h_create.restype = c_bool

    # disable tx API
    # bool k2h_disable_transaction(k2h_h handle)
    ret.k2h_disable_transaction.argtypes = [c_uint64]
    ret.k2h_disable_transaction.restype = c_bool

    # dump API
    # bool k2h_dump_head(k2h_h handle, FILE* stream)
    ret.k2h_dump_head.argtypes = [c_uint64, POINTER(FILE)]
    ret.k2h_dump_head.restype = c_bool
    # bool k2h_dump_keytable(k2h_h handle, FILE* stream)
    ret.k2h_dump_keytable.argtypes = [c_uint64, POINTER(FILE)]
    ret.k2h_dump_keytable.restype = c_bool
    # bool k2h_dump_full_keytable(k2h_h handle, FILE* stream)
    ret.k2h_dump_full_keytable.argtypes = [c_uint64, POINTER(FILE)]
    ret.k2h_dump_full_keytable.restype = c_bool
    # bool k2h_dump_elementtable(k2h_h handle, FILE* stream)
    ret.k2h_dump_elementtable.argtypes = [c_uint64, POINTER(FILE)]
    ret.k2h_dump_elementtable.restype = c_bool
    # bool k2h_dump_full(k2h_h handle, FILE* stream)
    ret.k2h_dump_full.argtypes = [c_uint64, POINTER(FILE)]
    ret.k2h_dump_full.restype = c_bool

    # get value API
    # char* k2h_get_str_direct_value_wp(k2h_h handle, const char* pkey, const char* pass)
    ret.k2h_get_str_direct_value_wp.argtypes = [c_uint64, c_char_p, c_char_p]
    ret.k2h_get_str_direct_value_wp.restype = c_char_p

    # get attrs API
    # PK2HATTRPCK k2h_get_direct_attrs(k2h_h handle, const unsigned char* pkey,
    # size_t keylength, int* pattrspckcnt)
    ret.k2h_get_direct_attrs.argtypes = [c_uint64, c_char_p, c_size_t, POINTER(c_int)]
    ret.k2h_get_direct_attrs.restype = POINTER(AttrPack)

    # get subkeys API
    # PK2HKEYPCK k2h_get_direct_subkeys(k2h_h handle, const unsigned char* pkey,
    # size_t keylength, int* pskeypckcnt)
    ret.k2h_get_direct_subkeys.argtypes = [c_uint64, c_char_p, c_size_t, POINTER(c_int)]
    ret.k2h_get_direct_subkeys.restype = POINTER(KeyPack)

    # get transaction API
    # int k2h_get_transaction_archive_fd(k2h_h handle)
    ret.k2h_get_transaction_archive_fd.argtypes = [c_uint64]
    ret.k2h_get_transaction_archive_fd.restype = c_int

    # int k2h_get_transaction_thread_pool(void)
    ret.k2h_get_transaction_thread_pool.argtypes = []
    ret.k2h_get_transaction_thread_pool.restype = c_int

    # load archive API
    # bool k2h_load_archive(k2h_h handle, const char* filepath, bool errskip)
    ret.k2h_load_archive.argtypes = [c_uint64, c_char_p, c_bool]
    ret.k2h_load_archive.restype = c_bool

    # open API
    ret.k2h_open_mem.restype = c_uint64
    ret.k2h_open_mem.argtypes = [c_int, c_int, c_int, c_int]
    ret.k2h_open.restype = c_uint64
    ret.k2h_open.argtypes = [
        c_char_p,
        c_bool,
        c_bool,
        c_bool,
        c_int,
        c_int,
        c_int,
        c_int,
    ]
    ret.k2h_open_rw.restype = c_uint64
    ret.k2h_open_rw.argtypes = [c_char_p, c_bool, c_int, c_int, c_int, c_int]
    ret.k2h_open_ro.restype = c_uint64
    ret.k2h_open_ro.argtypes = [c_char_p, c_bool, c_int, c_int, c_int, c_int]
    ret.k2h_open_tempfile.restype = c_uint64
    ret.k2h_open_tempfile.argtypes = [c_char_p, c_bool, c_int, c_int, c_int, c_int]

    # print API
    # bool k2h_print_attr_version(k2h_h handle, FILE* stream)
    ret.k2h_print_attr_version.argtypes = [c_uint64, POINTER(FILE)]
    ret.k2h_print_attr_version.restype = c_bool
    #
    # bool k2h_print_attr_information(k2h_h handle, FILE* stream)
    ret.k2h_print_attr_information.argtypes = [c_uint64, POINTER(FILE)]
    ret.k2h_print_attr_information.restype = c_bool
    #
    # bool k2h_print_state(k2h_h handle, FILE* stream)
    ret.k2h_print_state.argtypes = [c_uint64, POINTER(FILE)]
    ret.k2h_print_state.restype = c_bool

    # void k2h_print_version(FILE* stream)
    ret.k2h_print_version.argtypes = [POINTER(FILE)]
    ret.k2h_print_version.restype = None

    # put_archive API
    # bool k2h_put_archive(k2h_h handle, const char* filepath, bool errskip)
    ret.k2h_put_archive.argtypes = [c_uint64, c_char_p, c_bool]
    ret.k2h_put_archive.restype = c_bool

    # remove API
    # bool k2h_remove_str_all(k2h_h handle, const char* pkey)
    ret.k2h_remove_str_all.argtypes = [c_uint64, c_char_p]
    ret.k2h_remove_str_all.restype = c_bool
    # bool k2h_remove_str(k2h_h handle, const char* pkey)
    ret.k2h_remove_str.argtypes = [c_uint64, c_char_p]
    ret.k2h_remove_str.restype = c_bool

    # rename API
    # bool k2h_rename_str(k2h_h handle, const char* pkey, const char* pnewkey)
    ret.k2h_rename_str.argtypes = [c_uint64, c_char_p, c_char_p]
    ret.k2h_rename_str.restype = c_bool

    # remove subkey API
    # bool k2h_remove_str_subkey(k2h_h handle, const char* pkey, const char* psubkey)
    ret.k2h_remove_str_subkey.argtypes = [c_uint64, c_char_p, c_char_p]
    ret.k2h_remove_str_subkey.restype = c_bool

    # set_common_attr
    # bool k2h_set_common_attr(k2h_h handle, const bool* is_mtime, const bool* is_defenc,
    # const char* passfile, const bool* is_history, const c_ulong* expire)
    ret.k2h_set_common_attr.argtypes = [
        c_uint64,
        POINTER(c_bool),
        POINTER(c_bool),
        c_char_p,
        POINTER(c_bool),
        POINTER(c_ulong),
    ]
    ret.k2h_set_common_attr.restype = c_bool

    # set loglevel
    ret.k2h_set_debug_level_silent.argtypes = []
    ret.k2h_set_debug_level_silent.restype = None
    ret.k2h_set_debug_level_error.argtypes = []
    ret.k2h_set_debug_level_error.restype = None
    ret.k2h_set_debug_level_warning.argtypes = []
    ret.k2h_set_debug_level_warning.restype = None
    ret.k2h_set_debug_level_message.argtypes = []
    ret.k2h_set_debug_level_message.restype = None
    # set value
    # bool k2h_set_str_value_wa(k2h_h handle, const char* pkey, const char* pval, const char* pass,
    # const time_t* expire)
    ret.k2h_set_str_value_wa.argtypes = [
        c_uint64,
        c_char_p,
        c_char_p,
        c_char_p,
        POINTER(c_ulong),
    ]
    ret.k2h_set_str_value_wa.restype = c_bool

    # set transaction
    # bool k2h_transaction_param(k2h_h handle, bool enable, const char* transfile,
    # const unsigned char* pprefix, size_t prefixlen, const unsigned char* pparam, size_t paramlen)
    # bool k2h_transaction_param_we(k2h_h handle, bool enable, const char* transfile,
    # const unsigned char* pprefix, size_t prefixlen, const unsigned char* pparam, size_t paramlen,
    # const time_t* expire)
    ret.k2h_transaction_param_we.argtypes = [
        c_uint64,
        c_bool,
        c_char_p,
        c_char_p,
        c_size_t,
        c_char_p,
        c_size_t,
        POINTER(c_ulong),
    ]
    ret.k2h_transaction_param_we.restype = c_bool

    # bool k2h_set_transaction_thread_pool(int count)
    ret.k2h_set_transaction_thread_pool.argtypes = [c_int]
    ret.k2h_set_transaction_thread_pool.restype = c_int

    return ret


# Initializes library handlers
_init_library_handle()


# Gets library handler
def get_library_handle():
    """Gets C library handles"""
    return _init_library_handle()


# Configures logger using std logging. default puts to stderr in warning.
def _configure_logger(log_file="sys.stderr", log_level=logging.WARNING):
    LOG.setLevel(log_level)

    # 2. formatter
    formatter = logging.Formatter(
        "%(asctime)-15s %(levelname)s %(name)s:%(lineno)d %(message)s"
    )  # hardcoding

    # 3. log_file
    if log_file and isinstance(log_file, str):
        if log_file != "sys.stderr":
            # Add the log message handler to the logger
            handler = TimedRotatingFileHandler(
                log_file, when="midnight", encoding="UTF-8", backupCount=31
            )
            handler.setFormatter(formatter)
            LOG.addHandler(handler)
            return

    # default logger
    stream_handler = StreamHandler(sys.stderr)
    stream_handler.setFormatter(formatter)
    LOG.addHandler(stream_handler)


# Configures the loglevel.
def set_log_level(log_level):
    """Sets the log level"""
    LOG.setLevel(log_level)


# Initializes loggging handlers
_configure_logger()

#
# import k2hash modules
#
from k2hash.k2hash import K2hash, K2hashIterator  # noqa: pylint:disable=wrong-import-position
from k2hash.basequeue import BaseQueue  # noqa: pylint:disable=wrong-import-position
from k2hash.keyqueue import KeyQueue  # noqa: pylint:disable=wrong-import-position
from k2hash.queue import Queue  # noqa: pylint:disable=wrong-import-position

#
# Local variables:
# tab-width: 4
# c-basic-offset: 4
# End:
# vim600: expandtab sw=4 ts=4 fdm=marker
# vim<600: expandtab sw=4 ts=4
#



# -*- coding: utf-8 -*-
#
# K2hash Python Driver under MIT License
#
# Copyright (c) 2022 Yahoo Japan Corporation
#
# For the full copyright and license information, please view
# the license file that was distributed with this source code.
#
# AUTHOR:   Hirotaka Wakabayashi
# CREATE:   Tue Feb 08 2022
# REVISION:
#
"""K2hash Python Driver under MIT License"""
from __future__ import absolute_import

import logging
from ctypes import c_char_p, c_int, c_uint64, pointer

from k2hash import K2hash, BaseQueue

LOG = logging.getLogger(__name__)


class KeyQueue(BaseQueue):  # noqa: pylint: disable=too-many-instance-attributes
    """
    KeyQueue class provides methods to handle key/value pairs in k2hash hash database.
    """

    def __init__(  # noqa: pylint: disable=too-many-arguments
        self, k2h, fifo=True, prefix=None, password=None, expire_duration=None
    ):
        """
        Initialize a new KeyQueue instnace.
        """
        super().__init__(k2h, fifo=True, prefix=None, password=None, expire_duration=None)
        handle = self._libk2hash.k2h_keyq_handle_str_prefix(
            self._k2h_handle,
            self._fifo,
            (c_char_p(self._prefix.encode()) if self._prefix else None),
        )

        if handle == K2hash.K2H_INVALID_HANDLE:
            raise RuntimeError("handle should not be K2H_INVALID_HANDLE")
        self._handle = handle

    def put(self, obj):
        """Inserts an element into the tail of this queue."""
        if obj and not isinstance(obj, dict):
            raise TypeError("obj should be a dict object")
        if len(obj) == 0:
            raise ValueError("obj size should be greater than zero")
        for key, val in obj.items():
            if not isinstance(key, str):
                raise TypeError("key should be a str obj")
            if not isinstance(val, str):
                raise TypeError("val should be a str obj")

        for key, val in obj.items():
            res = self._libk2hash.k2h_keyq_str_push_keyval_wa(
                self._handle,
                c_char_p(key.encode()),
                c_char_p(val.encode()),
                (c_char_p(self._password.encode()) if self._password else None),
                (
                    pointer(c_uint64(self._expire_duration))
                    if self._expire_duration
                    else None
                ),
            )
            if res:
                LOG.debug("q_push:{%s}", res)
            else:
                return False
        return True

    def clear(self):
        """Removes all of the elements from this collection (optional operation)."""
        count = self.qsize()
        if count > 0:
            res = self._libk2hash.k2h_keyq_remove(self._handle, c_int(count))
            return res
        return True

    def close(self):
        """
        Free QueueHandle
        """
        res = self._libk2hash.k2h_keyq_free(self._handle)
        return res

    def qsize(self):
        """Returns the number of queue."""
        res = self._libk2hash.k2h_keyq_count(self._handle)
        return res

    def element(self, position=0):
        """Finds and gets a object from the head of this queue."""
        if not isinstance(position, int):
            raise TypeError("position should be a int object")
        if position < 0:
            raise ValueError("count should be positive")

        ppkey = pointer(c_char_p("".encode()))
        ppval = pointer(c_char_p("".encode()))
        res = self._libk2hash.k2h_keyq_str_read_keyval_wp(
            self._handle,
            ppkey,
            ppval,
            c_int(position),
            (c_char_p(self._password.encode()) if self._password else None),
        )
        res = {}
        if ppkey.contents.value:
            pkey = ppkey.contents.value.decode()
            self._libc.free(ppkey.contents)
            res[pkey] = ""
        if ppval.contents.value:
            pval = ppval.contents.value.decode()
            self._libc.free(ppval.contents)
            res[pkey] = pval
        return res

    def empty(self):
        """Returns true if, and only if, queue size is 0."""
        res = self._libk2hash.k2h_keyq_empty(self._handle)
        return res

    def get(self):
        """Finds and gets a object from the head of this queue."""
        ppkey = pointer(c_char_p("".encode()))
        ppval = pointer(c_char_p("".encode()))
        res = self._libk2hash.k2h_keyq_str_pop_keyval_wp(
            self._handle,
            ppkey,
            ppval,
            (c_char_p(self._password.encode()) if self._password else None),
        )
        res = {}
        if ppkey.contents.value:
            pkey = ppkey.contents.value.decode()
            self._libc.free(ppkey.contents)
            res[pkey] = ""
        if ppval.contents.value:
            pval = ppval.contents.value.decode()
            self._libc.free(ppval.contents)
            res[pkey] = pval
        return res

    def print(self):
        """Print the objects in this queue."""
        res = self._libk2hash.k2h_keyq_dump(self._handle, None)
        return res

    def remove(self, count=1):
        """Removes objects from this queue."""
        if not isinstance(count, int):
            raise TypeError("count should be a int object")
        if count <= 0:
            raise ValueError("count should be positive")

        vals = []
        for _ in range(count):
            val = self._libk2hash.k2h_keyq_remove(self._handle, 1)
            if val:
                vals.append(val)
        return vals


#
# Local variables:
# tab-width: 4
# c-basic-offset: 4
# End:
# vim600: expandtab sw=4 ts=4 fdm=marker
# vim<600: expandtab sw=4 ts=4
#



# -*- coding: utf-8 -*-
#
# K2hash Python Driver under MIT License
#
# Copyright (c) 2022 Yahoo Japan Corporation
#
# For the full copyright and license information, please view
# the license file that was distributed with this source code.
#
# AUTHOR:   Hirotaka Wakabayashi
# CREATE:   Tue Feb 08 2022
# REVISION:
#
"""K2hash Python Driver under MIT License"""
from __future__ import absolute_import

import logging

from k2hash import K2hash

LOG = logging.getLogger(__name__)


class BaseQueue:  # noqa: pylint: disable=too-many-instance-attributes
    """
    Baseueue class provides methods to handle key/value pairs in k2hash hash database.
    """

    def __init__(  # noqa: pylint: disable=too-many-arguments
        self, k2h, fifo=True, prefix=None, password=None, expire_duration=None
    ):
        """
        Initialize a new BaseQueue instnace.
        """
        if not isinstance(k2h, K2hash):
            raise TypeError("k2h should be a K2hash object")
        self._k2h_handle = k2h.handle
        self._libc = k2h.libc
        self._libk2hash = k2h.libk2hash

        if fifo and not isinstance(fifo, bool):
            raise TypeError("fifo should be a boolean object")
        self._fifo = fifo
        if prefix and not isinstance(prefix, str):
            raise TypeError("prefix should be a string object")
        self._prefix = prefix
        if password and not isinstance(password, str):
            raise TypeError("password should be a string object")
        self._password = password
        if expire_duration and not isinstance(expire_duration, int):
            raise TypeError("expire_duration should be a boolean object")
        if expire_duration and expire_duration <= 0:
            raise ValueError("expire_duration should not be positive")
        self._expire_duration = expire_duration

        # initializes self._handle, which should be set in subclasses
        self._handle = K2hash.K2H_INVALID_HANDLE

    @property
    def handle(self):
        """Returns a Queue handle."""
        return self._handle

    def __repr__(self):
        """Returns full of members as a string."""
        attrs = []
        for attr in [
            "_handle",
            "_k2h_handle",
            "_libk2hash",
            "_fifo",
            "_prefix",
            "_expire_duration",
        ]:  # should be hardcoded.
            val = getattr(self, attr)
            if val:
                attrs.append((attr, repr(val)))
            else:
                attrs.append((attr, ""))
            values = ", ".join(["%s=%s" % i for i in attrs])  # noqa: pylint:disable=consider-using-f-string
        return f"<_{self.__class__.__name__} " + values + ">"

#
# Local variables:
# tab-width: 4
# c-basic-offset: 4
# End:
# vim600: expandtab sw=4 ts=4 fdm=marker
# vim<600: expandtab sw=4 ts=4
#



# -*- coding: utf-8 -*-
#
# K2hash Python Driver under MIT License
#
# Copyright (c) 2022 Yahoo Japan Corporation
#
# For the full copyright and license information, please view
# the license file that was distributed with this source code.
#
# AUTHOR:   Hirotaka Wakabayashi
# CREATE:   Tue Feb 08 2022
# REVISION:
#
"""K2hash Python Driver under MIT License"""
from __future__ import absolute_import

import ctypes
import logging
import os
import sys
from ctypes import byref, c_bool, c_char_p, c_int, c_size_t, c_uint64, pointer
from pathlib import Path

import k2hash
from k2hash import DumpLevel, LogLevel, OpenFlag, TimeUnit

LOG = logging.getLogger(__name__)


class K2hashIterator:
    """implements iterator of k2hash"""

    def __init__(self, k2h, key=None):
        """Provides constructor"""
        if not isinstance(k2h, K2hash):
            raise TypeError("k2h should be a K2hash object")
        self._k2h_handle = k2h.handle
        self._libc = k2h.libc
        self._libk2hash = k2h.libk2hash
        self._key = None
        if key and isinstance(key, str):
            self._key = key
        if key and not isinstance(key, str):
            raise TypeError("key should be a str object")

        if self._key:
            handle = self._libk2hash.k2h_find_first_str_subkey(
                self._k2h_handle, (c_char_p(self._key.encode()))
            )
        else:
            handle = self._libk2hash.k2h_find_first(self._k2h_handle)

        if handle == K2hash.K2H_INVALID_HANDLE:
            raise RuntimeError("handle should not be K2H_INVALID_HANDLE")
        self._handle = handle

    def __iter__(self):
        """Implements iter() itrator interface"""
        return self

    def __next__(self):
        """Implements next() itrator interface"""
        ppkey = pointer(c_char_p("".encode()))
        pkeylength = pointer(c_size_t(0))
        res = self._libk2hash.k2h_find_get_key(self._handle, ppkey, pkeylength)
        if res:
            keylength = pkeylength.contents.value
            if keylength > 0:
                pkey = ppkey.contents.value.decode()
                self._libc.free(ppkey.contents)
                self._handle = self._libk2hash.k2h_find_next(self._handle)
                return pkey
        raise StopIteration


class K2hash:  # noqa: pylint: disable=too-many-instance-attributes,too-many-public-methods
    """
    K2hash class provides methods to handle key/value pairs in k2hash hash database.
    """

    K2H_INVALID_HANDLE = 0

    def get_iterator(self, key=None):
        """Returns the k2hash iterator"""
        if not self._iterator:
            self._iterator = K2hashIterator(self, key)
        return self._iterator

    def _set_k2h_handle(self):
        """Sets the k2h handle"""
        if self._k2hfile == "":
            if self._flag is None:
                # case1. no k2hfile + no flag ---> k2hash_open_memory
                handle = self._libk2hash.k2h_open_mem(
                    self._maskbit, self._cmaskbit, self._maxelementcnt, self._pagesize
                )
            elif self._flag == OpenFlag.MEMORY:
                # case2. no k2hfile + flag. k2hash_open_memory if flag is memory, otherwise error.
                handle = self._libk2hash.k2h_open_mem(
                    self._maskbit, self._cmaskbit, self._maxelementcnt, self._pagesize
                )
            else:
                raise ValueError("flag should be memory if k2hfile is empty")
        else:
            if self._flag is None:
                # case3. k2hfile + no flag ---> k2h_open
                handle = self._libk2hash.k2h_open(
                    c_char_p(self._k2hfile.encode()),
                    self._readonly,
                    self._removefile,
                    self._fullmap,
                    self._maskbit,
                    self._cmaskbit,
                    self._maxelementcnt,
                    self._pagesize,
                )
            elif self._flag == OpenFlag.EDIT:
                # case4. k2hfile + flag ---> k2hash_open_{rw|r|tempfile} if flag is not memory,
                # Otherwise error
                handle = self._libk2hash.k2h_open_rw(
                    c_char_p(self._k2hfile.encode()),
                    self._fullmap,
                    self._maskbit,
                    self._cmaskbit,
                    self._maxelementcnt,
                    self._pagesize,
                )
            elif self._flag == OpenFlag.READ:
                # Checks if self._k2hfile exists.
                if os.path.exists(Path(self._k2hfile)) is True:
                    handle = self._libk2hash.k2h_open_ro(
                        c_char_p(self._k2hfile.encode()),
                        self._fullmap,
                        self._maskbit,
                        self._cmaskbit,
                        self._maxelementcnt,
                        self._pagesize,
                    )
                else:
                    raise RuntimeError(f"{self._k2hfile} should exist")
            elif self._flag == OpenFlag.TEMPFILE:
                handle = self._libk2hash.k2h_open_tempfile(
                    c_char_p(self._k2hfile.encode()),
                    self._fullmap,
                    self._maskbit,
                    self._cmaskbit,
                    self._maxelementcnt,
                    self._pagesize,
                )
            else:
                raise ValueError("k2hfile should be empty if using flag is memory")

        if handle == self.__class__.K2H_INVALID_HANDLE:
            raise RuntimeError("handle should not be K2H_INVALID_HANDLE")
        self._handle = handle

    def __init__(  # noqa: pylint: disable=too-many-branches,too-many-arguments
        self,
        k2hfile="",
        flag=None,
        readonly=True,
        removefile=True,
        fullmap=True,
        maskbit=8,
        cmaskbit=4,
        maxelementcnt=1024,
        pagesize=512,
        waitms=0,
        logfile="",
    ):
        """
        Initialize a new K2hash instnace.
        """
        self._handle = 0
        self._iterator = None
        self._flag = None

        if not isinstance(k2hfile, str):
            raise TypeError("k2hfile should currently be a str object")
        self._k2hfile = k2hfile
        if flag and not isinstance(flag, OpenFlag):
            raise TypeError("flag should be a OpenFlag object")
        self._flag = flag
        if self._k2hfile == "":
            self._flag = OpenFlag.MEMORY  # overrides flag
        if not isinstance(readonly, bool):
            raise TypeError("readonly should be an bool object")
        self._readonly = readonly
        if not isinstance(removefile, bool):
            raise TypeError("removefile should be an bool object")
        self._removefile = removefile
        if not isinstance(fullmap, bool):
            raise TypeError("fullmap should be an bool object")
        self._fullmap = fullmap
        if not isinstance(maskbit, int):
            raise TypeError("maskbit should be an bool object")
        self._maskbit = maskbit
        if not isinstance(cmaskbit, int):
            raise TypeError("cmaskbit should be an bool object")
        self._cmaskbit = cmaskbit
        if not isinstance(maxelementcnt, int):
            raise TypeError("maxelementcnt should be an bool object")
        self._maxelementcnt = maxelementcnt
        if not isinstance(pagesize, int):
            raise TypeError("pagesize should be an bool object")
        self._pagesize = pagesize
        if not isinstance(waitms, int):
            raise TypeError("waitms should be an bool object")
        self._waitms = waitms
        if not isinstance(logfile, str):
            raise TypeError("logfile should currently be a str object")
        self._logfile = logfile

        try:
            # https://docs.python.org/3/library/ctypes.html#ctypes.LibraryLoader.LoadLibrary
            self._libc = k2hash.get_library_handle()["c"]
            self._libk2hash = k2hash.get_library_handle()["k2hash"]
            if not self._libk2hash:
                raise RuntimeError("unable to load k2hash library")
        except:
            LOG.error("Unexpected error:{%s}", sys.exc_info()[0])
            raise

        self._set_k2h_handle()

    @property
    def libk2hash(self):
        """returns libk2hash handle"""
        return self._libk2hash

    @property
    def libc(self):
        """returns libc handle"""
        return self._libc

    def set(  # noqa: pylint: disable=too-many-arguments
        self, key, val, password=None, expire_duration=None, time_unit=TimeUnit.SECONDS
    ):
        """Sets a key/value pair"""
        if not isinstance(key, str):
            raise TypeError("key should currently be a str object")
        if not key:
            raise ValueError("key should not be empty")
        if not isinstance(val, str):
            raise TypeError("val should currently be a str object")
        if password and not isinstance(password, str):
            raise TypeError("password should be a str object")
        if password and password == "":
            raise ValueError("password should not be empty")
        if expire_duration and not isinstance(expire_duration, int):
            raise TypeError("expire_duration should be a int object")
        if expire_duration and expire_duration <= 0:
            raise ValueError("expire_duration should not be positive")
        if time_unit and not isinstance(time_unit, TimeUnit):
            raise TypeError("time_unit should be a TimeUnit object")

        res = self._libk2hash.k2h_set_str_value_wa(
            self._handle,
            c_char_p(key.encode()),
            c_char_p(val.encode()),
            (c_char_p(password.encode()) if password else None),
            (pointer(c_uint64(expire_duration)) if expire_duration else None),
        )

        LOG.debug("ret:%s", res)
        return res

    def get(self, key, password=None):
        """Gets the value"""
        if not isinstance(key, str):
            raise TypeError("key should currently be a str object")
        if not key:
            raise ValueError("key should not be empty")
        if password and not isinstance(password, str):
            raise TypeError("password should be a str object")
        if password and password == "":
            raise ValueError("password should not be empty")

        val = self._libk2hash.k2h_get_str_direct_value_wp(
            self._handle,
            c_char_p(key.encode()),
            (c_char_p(password.encode()) if password else None),
        )

        if val:
            return val.decode()
        return ""

    def add_attribute_plugin_lib(self, path):
        """Adds a shared library that handles an attribute"""
        if not isinstance(path, str):
            raise TypeError("path should currently be a str object")
        if not os.path.exists(Path(path)):
            raise RuntimeError(f"path {path} should exists")

        res = self._libk2hash.k2h_add_attr_plugin_library(
            self._handle, c_char_p(path.encode())
        )

        if not res:
            LOG.error("error in k2h_add_attr_plugin_library")
        return res

    def add_decryption_password(self, password):
        """Adds a passphrase to decrypt a value."""
        if not isinstance(password, str):
            raise TypeError("key should currently be a str object")
        if password and password == "":
            raise ValueError("password should not be empty")

        res = self._libk2hash.k2h_add_attr_crypt_pass(
            self._handle, c_char_p(password.encode()), False
        )

        if not res:
            LOG.error("error in k2h_add_attr_crypt_pass")
        return res

    def add_subkey(  # noqa: pylint: disable=too-many-arguments
        self,
        key,
        subkey,
        subval,
        password=None,
        expire_duration=None,
        time_unit=TimeUnit.SECONDS,
    ):
        """Adds subkeys to a key/value pair."""
        if not isinstance(key, str):
            raise TypeError("key should currently be a str object")
        if not key:
            raise ValueError("key should not be empty")
        if not isinstance(subkey, str):
            raise TypeError("subkey should be a str object")
        if not subkey:
            raise ValueError("subkey should not be empty")
        if not isinstance(subval, str):
            raise TypeError("subval should be a str object")
        if password and not isinstance(password, str):
            raise TypeError("password should currently be a str object")
        if password and password == "":
            raise ValueError("password should not be empty")
        if expire_duration and not isinstance(expire_duration, int):
            raise TypeError("expire_duration should be a int object")
        if expire_duration and expire_duration <= 0:
            raise ValueError("expire_duration should not be positive")
        if time_unit and not isinstance(time_unit, TimeUnit):
            raise TypeError("time_unit should be a TimeUnit object")

        res = self._libk2hash.k2h_add_subkey_wa(
            self._handle,
            c_char_p(key.encode()),
            c_size_t(len(key) + 1),
            c_char_p(subkey.encode()),
            c_size_t(len(subkey) + 1),
            c_char_p(subval.encode()),
            c_size_t(len(subval) + 1),
            (c_char_p(password.encode()) if password else None),
            (pointer(c_uint64(expire_duration)) if expire_duration else None),
        )

        if not res:
            LOG.error("error in k2h_add_str_subkey_wa")
        return res

    def begin_tx(self, txfile, prefix=None, param=None, expire_duration=None):
        """Starts a transaction logging."""
        if not isinstance(txfile, str):
            raise TypeError("txfile should be a str object")
        if txfile == "":
            raise ValueError("txfile should not be empty")
        if prefix and not isinstance(prefix, str):
            raise TypeError("prefix should be a str object")
        if prefix and prefix == "":
            raise ValueError("prefix should not be empty")
        if param and not isinstance(param, str):
            raise TypeError("param should be a str object")
        if param and param == "":
            raise ValueError("param should not be empty")
        if expire_duration and not isinstance(expire_duration, int):
            raise TypeError("expire_duration should be a int object")
        if expire_duration and expire_duration <= 0:
            raise ValueError("expire_duration should not be positive")

        res = self._libk2hash.k2h_transaction_param_we(
            self._handle,
            True,
            c_char_p(txfile.encode()),
            (c_char_p(prefix.encode()) if prefix else None),
            (c_size_t(len(prefix)) if prefix else 0),
            (c_char_p(param.encode()) if param else None),
            (c_size_t(len(param)) if param else 0),
            (pointer(c_uint64(expire_duration)) if expire_duration else None),
        )

        if not res:
            LOG.error("error in k2h_transaction_param_we")
        return res

    def close(self):
        """Closes a k2h file."""
        res = self._libk2hash.k2h_close_wait(self._handle, self._waitms)
        if not res:
            LOG.error("error in k2h_close_wait")
        return res

    @staticmethod
    def create(  # noqa: pylint: disable=too-many-branches
        pathname, maskbit=8, cmaskbit=4, maxelementcnt=1024, pagesize=512
    ):
        """Creates a k2hash file."""
        if not isinstance(pathname, str):
            raise TypeError("pathname should be a str object")
        if not pathname:
            raise ValueError("pathname should not be empty")
        if not isinstance(maskbit, int):
            raise TypeError("maskbit should be an bool object")
        if maskbit <= 0:
            raise ValueError("maskbit should not be positive")
        if not isinstance(cmaskbit, int):
            raise TypeError("cmaskbit should be an bool object")
        if cmaskbit <= 0:
            raise ValueError("cmaskbit should not be positive")
        if not isinstance(maxelementcnt, int):
            raise TypeError("maxelementcnt should be an bool object")
        if maxelementcnt <= 0:
            raise ValueError("maxelementcnt should not be positive")
        if not isinstance(pagesize, int):
            raise TypeError("pagesize should be an bool object")
        if pagesize <= 0:
            raise ValueError("pagesize should not be positive")

        try:
            # https://docs.python.org/3/library/ctypes.html#ctypes.LibraryLoader.LoadLibrary
            libk2hash = k2hash.get_library_handle()["k2hash"]
            if not libk2hash:
                raise RuntimeError("unable to load k2hash library")
        except:
            LOG.error("Unexpected error:{%s}", sys.exc_info()[0])
            raise

        res = libk2hash.k2h_create(
            c_char_p(pathname.encode()),
            c_int(maskbit),
            c_int(cmaskbit),
            c_int(maxelementcnt),
            c_size_t(pagesize),
        )
        if not res:
            LOG.error("error in k2h_create")
        return res

    def dump_to_file(self, path, is_skip_error=True):
        """Dumps data to a file."""
        if not isinstance(path, str):
            raise TypeError("path should be a str object")
        if not path:
            raise ValueError("path should not be empty")
        if not isinstance(is_skip_error, bool):
            raise TypeError("is_skip_error should be a boolean object")

        res = self._libk2hash.k2h_put_archive(
            self._handle, c_char_p(path.encode()), is_skip_error
        )
        if not res:
            LOG.error("error in k2h_create")
        return res

    def enable_encryption(self, enable=True):
        """Enables a feature to encrypt a value."""
        if not isinstance(enable, bool):
            raise TypeError("enable should be a bool object")

        res = self._libk2hash.k2h_set_common_attr(
            self._handle, None, c_bool(enable), None, None, None
        )
        if not res:
            LOG.error("error in k2h_set_common_attr")
        return res

    def enable_history(self, enable=True):
        """Enables a feature to record a key modification history."""
        if not isinstance(enable, bool):
            raise TypeError("enable should be a bool object")

        res = self._libk2hash.k2h_set_common_attr(
            self._handle, None, None, None, c_bool(enable), None
        )
        if not res:
            LOG.error("error in k2h_set_common_attr")
        return res

    def enable_mtime(self, enable=True):
        """Enables a feature to record value modification time."""
        if not isinstance(enable, bool):
            raise TypeError("enable should be a bool object")

        res = self._libk2hash.k2h_set_common_attr(
            self._handle, c_bool(enable), None, None, None, None
        )
        if not res:
            LOG.error("error in k2h_set_common_attr")
        return res

    def get_attributes(self, key, use_str=True):
        """Gets attributes of a key."""
        if not isinstance(key, str):
            raise TypeError("key should currently be a str object")
        if not key:
            raise ValueError("key should not be empty")
        pattrspckcnt = c_int()
        res = self._libk2hash.k2h_get_direct_attrs(
            self._handle,
            c_char_p(key.encode()),
            c_size_t(len(key)),
            byref(pattrspckcnt),
        )
        LOG.debug(
            "type(res):{%s} pattrspckcnt.value{%s}", type(res), pattrspckcnt.value
        )
        attrs = {}
        for i in range(pattrspckcnt.value):
            key_buf = ctypes.create_string_buffer(res[i].keylength)
            val_buf = ctypes.create_string_buffer(res[i].vallength)
            for j in range(res[i].keylength):
                key_buf[j] = res[i].pkey[j]
            for k in range(res[i].vallength):
                val_buf[k] = res[i].pval[k]
            if use_str:
                attrs[key_buf.value.decode()] = val_buf.value.decode()
            else:
                attrs[key_buf.value] = val_buf.value
        return attrs

    @property
    def handle(self):
        """Returns a Queue handle."""
        return self._handle

    def get_subkeys(self, key, use_str=True):
        """Gets keys of subkeys of a key."""
        if not isinstance(key, str):
            raise TypeError("key should currently be a str object")
        if not key:
            raise ValueError("key should not be empty")

        pskeypckcnt = c_int()
        res = self._libk2hash.k2h_get_direct_subkeys(
            self._handle,
            c_char_p(key.encode()),
            c_size_t(len(key) + 1),
            byref(pskeypckcnt),
        )
        LOG.debug("%s", pskeypckcnt.value)
        subkeys = []
        for i in range(pskeypckcnt.value):
            buf = ctypes.create_string_buffer(res[i].length)
            for j in range(res[i].length):
                buf[j] = res[i].pkey[j]
            if use_str:
                subkeys.append(buf.value.decode())
            else:
                subkeys.append(buf.value)
        return subkeys

    def get_tx_file_fd(self):
        """Gets a transaction log file descriptor."""
        res = self._libk2hash.k2h_get_transaction_archive_fd(self._handle)
        return res

    @staticmethod
    def get_tx_pool_size():
        """Gets the number of transaction thread pool."""
        try:
            # https://docs.python.org/3/library/ctypes.html#ctypes.LibraryLoader.LoadLibrary
            libk2hash = k2hash.get_library_handle()["k2hash"]
            if not libk2hash:
                raise RuntimeError("unable to load k2hash library")
        except:
            LOG.error("Unexpected error:{%s}", sys.exc_info()[0])

            raise

        res = libk2hash.k2h_get_transaction_thread_pool()
        return res

    def load_from_file(self, path, is_skip_error=True):
        """Loads data from a file."""
        if not isinstance(path, str):
            raise TypeError("path should currently be a str object")
        if not os.path.exists(Path(path)):
            raise RuntimeError(f"path{path} should exists")
        if not isinstance(is_skip_error, bool):
            raise TypeError("is_skip_error should be a boolean object")

        res = self._libk2hash.k2h_load_archive(
            self._handle, c_char_p(path.encode()), c_bool(is_skip_error)
        )
        if not res:
            LOG.error("error in k2h_load_archive")
        return res

    def print_attribute_plugins(self):
        """Prints attribute plugins to stderr."""
        res = self._libk2hash.k2h_print_attr_version(self._handle, None)
        return res

    def print_attributes(self):
        """Prints attributes to stderr."""
        res = self._libk2hash.k2h_print_attr_information(self._handle, None)
        return res

    def print_data_stats(self):
        """Prints data statistics."""
        res = self._libk2hash.k2h_print_state(self._handle, None)
        return res

    def print_table_stats(self, level=DumpLevel.HEADER):
        """Prints k2hash key table information."""
        if not isinstance(level, DumpLevel):
            raise TypeError("level should be a DumpLevel object")

        if level == DumpLevel.HEADER:
            res = self._libk2hash.k2h_dump_head(self._handle, None)
        elif level == DumpLevel.HASH_TABLE:
            res = self._libk2hash.k2h_dump_keytable(self._handle, None)
        elif level == DumpLevel.SUB_HASH_TABLE:
            res = self._libk2hash.k2h_dump_full_keytable(self._handle, None)
        elif level == DumpLevel.ELEMENT:
            res = self._libk2hash.k2h_dump_elementtable(self._handle, None)
        elif level == DumpLevel.PAGE:
            res = self._libk2hash.k2h_dump_full(self._handle, None)
        else:
            raise ValueError(
                "level should be either HEADER, HASH_TABLE, SUB_HASH_TABLE, ELEMENT or PAGE"
            )
        return res

    def remove(self, key, remove_all_subkeys=False):
        """Removes a key."""
        if not isinstance(key, str):
            raise TypeError("key should be a str object")
        if not key:
            raise ValueError("key should not be empty")
        if not isinstance(remove_all_subkeys, bool):
            raise TypeError("remove_all_subkeys should be a boolean object")
        if remove_all_subkeys:
            res = self._libk2hash.k2h_remove_str_all(
                self._handle, c_char_p(key.encode())
            )
        else:
            res = self._libk2hash.k2h_remove_str(self._handle, c_char_p(key.encode()))
        return res

    def remove_subkeys(self, key, subkeys):
        """Removes subkeys from the key."""
        if not isinstance(key, str):
            raise TypeError("key should be a str object")
        if not key:
            raise ValueError("key should not be empty")
        if not isinstance(subkeys, list):
            raise TypeError("subkeys should be a list object")
        if not subkeys:
            raise ValueError("subkey should not be empty")
        for subkey in subkeys:
            if not isinstance(subkey, str):
                LOG.warning("subkey should be a str object")
                continue
            if not subkey:
                LOG.warning("subkey should not be empty")
                continue
            res = self._libk2hash.k2h_remove_str_subkey(
                self._handle, c_char_p(key.encode()), c_char_p(subkey.encode())
            )
            if not res:
                return False
        return True

    def rename(self, key, newkey):
        """Renames a key with a new key."""
        if not isinstance(key, str):
            raise TypeError("key should be a str object")
        if not key:
            raise ValueError("key should not be empty")
        if not isinstance(newkey, str):
            raise TypeError("newkey should be a str object")
        if not newkey:
            raise ValueError("newkey should not be empty")
        res = self._libk2hash.k2h_rename_str(
            self._handle, c_char_p(key.encode()), c_char_p(newkey.encode())
        )
        return res

    def set_attribute(self, key, attr_name, attr_val):
        """Sets an attribute of a key."""
        if not isinstance(key, str):
            raise TypeError("key should be a str object")
        if not key:
            raise ValueError("key should not be empty")
        if not isinstance(attr_name, str):
            raise TypeError("attr_name should be a str object")
        if not attr_name:
            raise ValueError("attr_name should not be empty")
        if not isinstance(attr_val, str):
            raise TypeError("attr_val should be a str object")
        if not attr_val:
            raise ValueError("attr_val should not be empty")
        res = self._libk2hash.k2h_add_attr(
            self._handle,
            c_char_p(key.encode()),
            c_size_t(len(key)),
            c_char_p(attr_name.encode()),
            c_size_t(len(attr_name) + 1),
            c_char_p(attr_val.encode()),
            c_size_t(len(attr_val) + 1),
        )
        return res

    def set_default_encryption_password(self, password):
        """Sets the default encryption passphrase."""
        if not isinstance(password, str):
            raise TypeError("password should be a str object")
        if not password:
            raise ValueError("password should not be empty")
        res = self._libk2hash.k2h_add_attr_crypt_pass(
            self._handle, c_char_p(password.encode()), c_bool(True)
        )
        return res

    def set_encryption_password_file(self, path):
        """Sets the data encryption password file."""
        if not isinstance(path, str):
            raise TypeError("path should currently be a str object")
        if not os.path.exists(Path(path)):
            raise RuntimeError(f"path{path} should exists")
        res = self._libk2hash.k2h_set_common_attr(
            self._handle, None, None, c_char_p(path.encode()), None, None
        )
        if not res:
            LOG.error("error in k2h_set_common_attr")
        return res

    def set_expiration_duration(self, expire_duration, time_unit=TimeUnit.SECONDS):
        """Sets the duration to expire a value."""
        if not isinstance(expire_duration, int):
            raise TypeError("expire_duration should be a int object")
        if expire_duration <= 0:
            raise ValueError("expire_duration should not be positive")
        if not isinstance(time_unit, TimeUnit):
            raise TypeError("time_unit should be a TimeUnit object")
        res = self._libk2hash.k2h_set_common_attr(
            self._handle, None, None, None, None, pointer(c_uint64(expire_duration))
        )
        if not res:
            LOG.error("error in k2h_set_common_attr")
        return res

    def set_log_level(self, level=LogLevel.INFO):
        """Creates a k2hash file."""
        if not isinstance(level, LogLevel):
            raise TypeError("level should be a LogLevel object")
        if not level:
            raise ValueError("level should not be empty")

        if level == LogLevel.SILENT:
            self._libk2hash.k2h_set_debug_level_silent()
        elif level == LogLevel.ERROR:
            self._libk2hash.k2h_set_debug_level_error()
        elif level == LogLevel.WARNING:
            self._libk2hash.k2h_set_debug_level_warning()
        elif level == LogLevel.INFO:
            self._libk2hash.k2h_set_debug_level_message()
        elif level == LogLevel.DEBUG:
            self._libk2hash.k2h_set_debug_level_dump()
        else:
            raise ValueError(
                "level should be either SILENT, ERROR, WARN, INFO or DEBUG"
            )

    def set_subkeys(  # noqa: pylint: disable=too-many-branches,too-many-arguments
        self,
        key,
        subkeys,
        password=None,
        expire_duration=None,
        time_unit=TimeUnit.SECONDS,
    ):
        """Sets subkeys."""
        if not isinstance(key, str):
            raise TypeError("key should be a str object")
        if not key:
            raise ValueError("key should not be empty")
        if not isinstance(subkeys, dict):
            raise TypeError("subkeys should be a dict object")
        if not subkeys:
            raise ValueError("subkey should not be empty")
        if password and not isinstance(password, str):
            raise TypeError("password should currently be a str object")
        if password and password == "":
            raise ValueError("password should not be empty")
        if expire_duration and not isinstance(expire_duration, int):
            raise TypeError("expire_duration should be a int object")
        if expire_duration and expire_duration <= 0:
            raise ValueError("expire_duration should not be positive")
        if time_unit and not isinstance(time_unit, TimeUnit):
            raise TypeError("time_unit should be a TimeUnit object")
        for subkey, subval in subkeys.items():
            if not isinstance(subkey, str):
                LOG.warning("subkey should be a str object")
                continue
            if not subkey:
                LOG.warning("subkey should not be empty")
                continue
            if not isinstance(subval, str):
                LOG.warning("subval should be a str object")
                continue
            if not subval:
                LOG.warning("subval should not be empty")
                continue
            res = self._libk2hash.k2h_add_subkey_wa(
                self._handle,
                c_char_p(key.encode()),
                c_size_t(len(key) + 1),
                c_char_p(subkey.encode()),
                c_size_t(len(subkey) + 1),
                c_char_p(subval.encode()),
                c_size_t(len(subval) + 1),
                (c_char_p(password.encode()) if password else None),
                (pointer(c_uint64(expire_duration)) if expire_duration else None),
            )
            if res:
                LOG.debug("k2h_add_str_subkeys:{%s}", res)
            else:
                return False
        return True

    @staticmethod
    def set_tx_pool_size(size):
        """Sets the number of transaction thread pool."""
        if not isinstance(size, int):
            raise TypeError("size should be a int object")
        if size < 0:
            raise ValueError("size should be positive")
        try:
            # https://docs.python.org/3/library/ctypes.html#ctypes.LibraryLoader.LoadLibrary
            libk2hash = k2hash.get_library_handle()["k2hash"]
            if not libk2hash:
                raise RuntimeError("unable to load k2hash library")
        except:
            LOG.error("Unexpected error:{%s}", sys.exc_info()[0])
            raise

        res = libk2hash.k2h_set_transaction_thread_pool(size)
        if not res:
            LOG.error("error in k2h_set_transaction_thread_pool")
        return res

    def stop_tx(self):
        """Stops a transaction logging."""
        res = self._libk2hash.k2h_disable_transaction(self._handle)
        if not res:
            LOG.error("error in k2h_disable_transaction")
        return res

    def __repr__(self):
        """Returns full of members as a string."""
        attrs = []
        for attr in [
            "_k2hfile",
            "_flag",
            "_readonly",
            "_removefile",
            "_fullmap",
            "_maskbit",
            "_cmaskbit",
            "_maxelementcnt",
            "_pagesize",
            "_waitms",
            "_logfile",
            "_libc",
            "_libk2hash",
            "_handle",
        ]:  # should be hardcoded.
            val = getattr(self, attr)
            if val:
                attrs.append((attr, repr(val)))
            else:
                attrs.append((attr, ""))
            values = ", ".join(["%s=%s" % i for i in attrs])  # noqa: pylint: disable=consider-using-f-string
        return "<_K2hash " + values + ">"

    @staticmethod
    def version():
        """Prints version information."""
        try:
            # https://docs.python.org/3/library/ctypes.html#ctypes.LibraryLoader.LoadLibrary
            libk2hash = k2hash.get_library_handle()["k2hash"]
            if not libk2hash:
                raise RuntimeError("unable to load k2hash library")
        except:
            LOG.error("Unexpected error:{%s}", sys.exc_info()[0])
            raise

        libk2hash.k2h_print_version(None)


#
# Local variables:
# tab-width: 4
# c-basic-offset: 4
# End:
# vim600: expandtab sw=4 ts=4 fdm=marker
# vim<600: expandtab sw=4 ts=4
#



# -*- coding: utf-8 -*-
#
# K2hash Python Driver under MIT License
#
# Copyright (c) 2022 Yahoo Japan Corporation
#
# For the full copyright and license information, please view
# the license file that was distributed with this source code.
#
# AUTHOR:   Hirotaka Wakabayashi
# CREATE:   Tue Feb 08 2022
# REVISION:
#
"""K2hash Python Driver under MIT License"""
from __future__ import absolute_import

import copy
import logging
from ctypes import (POINTER, c_char_p, c_int, c_size_t, c_ubyte, c_uint64,
                    cast, pointer)

from k2hash import AttrPack, K2hash, BaseQueue

LOG = logging.getLogger(__name__)


class Queue(BaseQueue):  # noqa: pylint:disable=too-many-instance-attributes
    """
    Queue class provides methods to handle key/value pairs in k2hash hash database.
    """

    def __init__(  # noqa: pylint: disable=too-many-arguments
        self, k2h, fifo=True, prefix=None, password=None, expire_duration=None
    ):
        """
        Initialize a new Queue instnace.
        """
        super().__init__(k2h, fifo=True, prefix=None, password=None, expire_duration=None)
        handle = self._libk2hash.k2h_q_handle_str_prefix(
            self._k2h_handle,
            self._fifo,
            (c_char_p(self._prefix.encode()) if self._prefix else None),
        )

        if handle == K2hash.K2H_INVALID_HANDLE:
            raise RuntimeError("handle should not be K2H_INVALID_HANDLE")
        self._handle = handle


    def put(self, obj, attrs=None):
        """Inserts an element into the tail of this queue."""
        value = []
        if not isinstance(obj, list) and not isinstance(obj, str):
            raise TypeError("obj should be a str or list object")
        if not obj:
            raise ValueError("obj should not be empty")
        if isinstance(obj, list):
            value += copy.deepcopy(obj)
        elif isinstance(obj, str):
            value.append(obj)

        if attrs and not isinstance(attrs, dict):
            raise TypeError("attrs should be a dict")
        ap_array_pointer = None
        ap_array = None
        if attrs:
            attrpack_len = len(attrs)
            ap_array = AttrPack * attrpack_len
            ap_list = []
            for key, val in attrs.items():
                attr_pack = AttrPack()
                key_bin = key.encode()
                attr_pack.pkey = cast(key_bin, POINTER(c_ubyte))
                attr_pack.keylength = c_size_t(len(key_bin))
                val_bin = val.encode()
                attr_pack.pval = cast(val_bin, POINTER(c_ubyte))
                attr_pack.vallength = c_size_t(len(val_bin))
                ap_list.append(attr_pack)
            ap_array_pointer = pointer(ap_array()[0])

        res = self._libk2hash.k2h_q_str_push_wa(
            self._handle,
            c_char_p(obj.encode()),
            (ap_array_pointer if ap_array else None),
            0,
            (c_char_p(self._password.encode()) if self._password else None),
            (
                pointer(c_uint64(self._expire_duration))
                if self._expire_duration
                else None
            ),
        )
        return res

    def clear(self):
        """Removes all of the elements from this collection (optional operation)."""
        count = self.qsize()
        if count > 0:
            res = self._libk2hash.k2h_q_remove(self._handle, c_int(count))
            return res
        return True

    def close(self):
        """
        Free QueueHandle
        """
        res = self._libk2hash.k2h_q_free(self._handle)
        return res

    def qsize(self):
        """Returns the number of queue."""
        res = self._libk2hash.k2h_q_count(self._handle)
        return res

    def element(self, position=0):
        """Finds and gets a object from the head of this queue."""
        if not isinstance(position, int):
            raise TypeError("position should be a int object")
        if position < 0:
            raise ValueError("count should be positive")

        ppdata = pointer(c_char_p("".encode()))
        pdatalen = pointer(c_size_t(0))
        self._libk2hash.k2h_q_read_wp(
            self._handle,
            ppdata,
            pdatalen,
            c_int(position),
            (c_char_p(self._password.encode()) if self._password else None),
        )
        datalen = pdatalen.contents.value
        if datalen > 0:
            pdata = ppdata.contents.value.decode()
            self._libc.free(ppdata.contents)
            return pdata
        return ""

    def empty(self):
        """Returns true if, and only if, queue size is 0."""
        res = self._libk2hash.k2h_q_empty(self._handle)
        return res

    def get(self):
        """Finds and gets a object from the head of this queue."""
        ppval = pointer(c_char_p("".encode()))
        res = self._libk2hash.k2h_q_str_pop_wp(
            self._handle,
            ppval,
            (c_char_p(self._password.encode()) if self._password else None),
        )

        if res and ppval.contents.value:
            pval = ppval.contents.value.decode()
            if ppval.contents:
                self._libc.free(ppval.contents)
            return pval
        return ""

    def print(self):
        """Print the objects in this queue."""
        res = self._libk2hash.k2h_q_dump(self._handle, None)
        return res

    def remove(self, count=1):
        """Removes objects from this queue."""
        if not isinstance(count, int):
            raise TypeError("count should be a int object")
        if count <= 0:
            raise ValueError("count should be positive")

        vals = []
        for _ in range(count):
            val = self._libk2hash.k2h_q_remove(self._handle, 1)
            if val:
                vals.append(val)
        return vals

#
# Local variables:
# tab-width: 4
# c-basic-offset: 4
# End:
# vim600: expandtab sw=4 ts=4 fdm=marker
# vim<600: expandtab sw=4 ts=4
#
